
module KeyShot12

  @PLATFORM_IS_OSX = (Object::RUBY_PLATFORM =~ /darwin/i) ? true : false
  @@LIB_NAME = 'KeyShot12Export'
  def self.PLATFORM_IS_OSX
    @PLATFORM_IS_OSX
  end
  
  @@KEYSHOT12_FOUND = false
  @@TEXTURE_PATH = ""
    
  def self.TEXTURE_PATH
    @@TEXTURE_PATH
  end
  
  require 'rexml/document'
  require 'json'

  if @PLATFORM_IS_OSX
    if Sketchup.version.to_f < 17
      require 'luxion_keyshot12/mac/KeyShot12Export.64_2.0.0'
    elsif Sketchup.version.to_f < 19
      require 'luxion_keyshot12/mac/KeyShot12Export.64_2.2.0'
    elsif Sketchup.version.to_f < 19.3
      require 'luxion_keyshot12/mac/KeyShot12Export.64_2.5.1'
    elsif Sketchup.version.to_f < 21.0
      require 'luxion_keyshot12/mac/KeyShot12Export.64_2.5.5'
    else
      require 'luxion_keyshot12/mac/KeyShot12Export.64_2.7.1'
    end
  else
    require 'win32/registry'

    module Win32::Registry::Constants
      KEY_WOW64_64KEY = 0x0100
      KEY_WOW64_32KEY = 0x0200
    end

    if Sketchup.version.to_f > 14
      arch = '32'
      if Sketchup.is_64bit?
        arch = '64'
      end
      if Sketchup.version.to_f < 17
        require 'luxion_keyshot12/win/KeyShot12Export.' + arch + '_2.0.0'
      elsif Sketchup.version.to_f < 19
        require 'luxion_keyshot12/win/KeyShot12Export.' + arch + '_2.2.0'
      elsif Sketchup.version.to_f < 19.3
        require 'luxion_keyshot12/win/KeyShot12Export.' + arch + '_2.5.1'
      elsif Sketchup.version.to_f < 21.0
        require 'luxion_keyshot12/win/KeyShot12Export.' + arch + '_2.5.5'
      else
        require 'luxion_keyshot12/win/KeyShot12Export.' + arch + '_2.7.1'
      end
    end
  end

  # Shim for the Set class which was moved in SketchUp 2014
  if defined?(Sketchup::Set)
    Set = Sketchup::Set
  end 

  class Options
    include KeyShot12
    attr_reader :export_camera
    attr_reader :double_sided_material
    attr_reader :export_images
    attr_reader :overwrite_textures

    def initialize
      if File.exists?(File.expand_path(File.dirname(__FILE__)) + '/options.xml')
        file = File.open(File.expand_path(File.dirname(__FILE__)) + '/options.xml')
        xml = file.read
        doc = REXML::Document.new(xml)
        @double_sided_material = (doc.elements["keyShotExporter/double_sided_material"].get_text()  == '1')
        @export_camera = (doc.elements["keyShotExporter/export_camera"].get_text()  == '1')
        @export_images = (doc.elements["keyShotExporter/export_images"].get_text()  == '1')
        @overwrite_textures = (doc.elements["keyShotExporter/overwrite_textures"].get_text()  == '1')
      else 
        @double_sided_material = false
        @export_camera = true
        @export_images = true
        @overwrite_textures = false
      end
    end
  end

  Struct.new("KeyshotMaterial", :name, :texture, :rgb, :alpha, :width, :height)
    
  class KeyShot12Exporter  
    include KeyShot12
    attr_accessor :ext_exporter
  
    def initialize
      @texture_set = Set.new
      @error = false
    end
  
    #Calls the argument method in a begin rescue clause, and set error condition if applicable. 
    def callCFunction(method, *args)
      begin
        method.call(*args)
      rescue => error
        UI.messagebox(error) 
        @error = true
        return
      end
    end

    #returns material struct
    def handleMaterial(input_material, parent)
      return nil if @error
      unless input_material 
        return nil
      end
      material = Struct::KeyshotMaterial.new("Default", "", [255, 255, 255], 1.0, 1, 1)

      materialname = ""
      texturename = ""
      width = 1
      height = 1
      
      material["name"] = input_material.display_name
      material["alpha"] = input_material.alpha
      material["rgb"] = input_material.color.to_a if input_material.color
      if input_material.texture
        if parent
          width = 1.0 / input_material.texture.width
          height = 1.0 / input_material.texture.height
        end
        material["texture"] = handleTexture(input_material.texture.filename, input_material, parent)
      end

      return material
    end

    #returns full texture path
    def handleTexture(input_name, material, parent)
      return nil if @error
      texturename = ''
      if File.exists?(input_name)
        texturename = input_name
      else
        texturename = input_name
        if texturename["/"]
          texturename = texturename[texturename.rindex("/")+1..-1]
        elsif texturename["\\"]
          texturename = texturename[texturename.rindex("\\")+1..-1]
        end
        # Texture writer converts .tga to .png - a pretty poorly documented feature
        texturename[texturename.size-3..-1] = "png" if texturename[texturename.size-3..-1] == "tga"  
        modelPath = Sketchup.active_model.path
        filename = 'untitled_scene'
        if File.exists?(modelPath)
          filename = File.basename(modelPath, '.*')
        end
        if TEXTURE_PATH != ""
          texturename = TEXTURE_PATH + "/" + filename + "/" + texturename
          dirname = TEXTURE_PATH + "/" + filename + "/"
          unless File.exists?(dirname)
            Dir.mkdir(dirname)
          end
        else
          texturename = input_name
        end
        if File.exists?(texturename) and (not @options.overwrite_textures)
          return texturename
        end

        unless File.exists?(texturename) and @texture_set.contains? texturename
          @texture_set.insert texturename
          
          # We first try to assign the material to an empty group and use that to export the texture
          # This is done to export the texture without transformations
          
          model = Sketchup.active_model
          entities = model.active_entities
          tmp = entities.add_group

          tmp.material = material
          twtmp = Sketchup.create_texture_writer
          twtmp.load(tmp)
          ret = twtmp.write(tmp, texturename)

          if ret != FILE_WRITE_OK
            # Some textures are missing an extension so try to export as jpg
            texturenameJPG = texturename + '.jpg'
            unless File.exists?(texturenameJPG) and @texture_set.contains? texturenameJPG
              ret = twtmp.write(tmp, texturenameJPG)
            end
          end
          entities.erase_entities([tmp])
          
          if ret != FILE_WRITE_OK
            begin
              # If the texture still isn't correctly exported we try to export it directly from the entity with the material
              # This sometimes seems to be necessary for Image entities. See KEYS-8733 for an example
              twtmp.load(parent)
              ret = twtmp.write(parent, texturename)
            rescue => error
              puts error
            end
          end          
          
        end
      end
      return texturename
    end

    def handleFace(entity, acc_trans, parent_mat, layer_name)
      return nil if @error
      if layer_name != ""
        if entity.layer.name != layer_name
          return
        end
      end

      frontMaterial = handleMaterial(entity.material, false)
      backMaterial = handleMaterial(entity.back_material, false)
      parentMaterial = handleMaterial(parent_mat, true)

      material = nil
      frontmat = true
      
      if frontMaterial
        material = frontMaterial
      else
        if backMaterial
          material = backMaterial
          frontmat = false
        else
          if parentMaterial
            material = parentMaterial
            useParentMat = true
          end    
        end
      end
          
      # tell sketchup to make a triangle mesh from the current face
      # 5 as argument makes the mesh include all uvs and normals
      # 6 as argument makes the mesh include back face uvs instead. 
      if frontmat
        mesh = entity.mesh 5
      else
        mesh = entity.mesh 6
      end
       
      # array of floats representing normal vectors
      narr = []
      i = 0
      (1..mesh.count_points).each { |i| narr = narr + mesh.normal_at(i).to_a}
      # call the export method in the dll
      if frontMaterial == backMaterial
        backMaterial = nil
      end
      callCFunction(SUKeyShotExport.method(:add_faces), mesh.points.length, mesh.polygons.length, mesh.points, mesh.polygons, narr, mesh.uvs(frontmat), frontMaterial, backMaterial, parentMaterial)
    end

    # Visit an entity passing face information to the dll and recursing on groups and components
    def visit(entity, acc_trans, parent_mat, layer_name)
      return nil if @error
      useParentMat = false
      if not entity.hidden?
        if entity.is_a?(Sketchup::Face)
          handleFace(entity, acc_trans, parent_mat, layer_name)
        end
        
        if entity.is_a?(Sketchup::Image) and @options.export_images
          width = entity.width
          height = entity.height
          trans = entity.transformation
          texturename = handleTexture(entity.path, entity.material, entity)
          callCFunction(SUKeyShotExport.method(:add_image), width, height, texturename, trans.to_a)
        end
        if entity.is_a?(Sketchup::ComponentInstance)
          if layer_name != ""
            if entity.layer.name != layer_name
              return
            end
          end
          transformation = entity.transformation
          name = ""
          if entity.name != ""
            name = entity.name
          else
            if entity.definition.name != ""
              name = entity.definition.name
            else
              if entity.material
                name = entity.material.display_name
              end
            end
          end
          callCFunction(SUKeyShotExport.method(:begin_group), name);
          callCFunction(SUKeyShotExport.method(:transform), transformation.to_a, entity.entityID)
          entity.definition.entities.each do |e|
            material = 0
            if entity.material
              material = entity.material
            else
              material = parent_mat
            end
            visit(e, transformation, material, "")
          end
          callCFunction(SUKeyShotExport.method(:end_group))
        end
        if entity.is_a?(Sketchup::Group)
          if layer_name != ""
            if entity.layer.name != layer_name
              return
            end
          end
          transformation = entity.transformation
          name = ""
          if entity.name != ""
            name = entity.name
          else
            if entity.material
              name = entity.material.display_name
            end
          end

          callCFunction(SUKeyShotExport.method(:begin_group), name)
          callCFunction(SUKeyShotExport.method(:transform), transformation.to_a, entity.entityID)
          entity.entities.each do |e|
            material = 0
            if entity.material
              material = entity.material
            else
              material = parent_mat
            end
            visit(e, transformation, material, "")
          end
          callCFunction(SUKeyShotExport.method(:end_group))
        end
      end
    end
  
  # Visit an entity passing face information to the dll and recursing on groups and components
    def visitAnimation(entity, acc_trans, layer_name)
      return nil if @error
      useParentMat = false
      if not entity.hidden?       
        if entity.is_a?(Sketchup::Group)
          if layer_name != ""
            if entity.layer.name != layer_name
              return
            end
          end
          transformation = entity.transformation
          callCFunction(SUKeyShotExport.method(:animation), transformation.to_a, entity.entityID)
          entity.entities.each do |e|
            visitAnimation(e, transformation, "")
          end
        end
        if entity.is_a?(Sketchup::ComponentInstance)
          if layer_name != ""
            if entity.layer.name != layer_name
              return
            end
          end
          transformation = entity.transformation
          callCFunction(SUKeyShotExport.method(:animation), transformation.to_a, entity.entityID)
          entity.definition.entities.each do |e|
            visitAnimation(e, transformation, "")
          end
        end
      end
    end
  
    def ks12export(location, cameraOnly)    
      @error = false  
      @texture_set.clear
      @options = Options.new
      Sketchup.active_model.start_operation("task",true)
      model = Sketchup.active_model
    
      callCFunction(SUKeyShotExport.method(:set_options), @options)

      # File management
      filename = ""
      unless model.title == ''
        filename += model.title + '.bip'
      else
        filename += 'untitled.bip'
      end

      if(location != "")
        save = true
      else
        save = false
      end
      
      puts 'Try connecting'
      callCFunction(SUKeyShotExport.method(:open_file), filename, location, save, cameraOnly, Sketchup.version)
      
      if @error
        model.commit_operation
        return
      end
      puts 'connection established'
      camera = model.active_view.camera

      position = camera.eye.to_a
      direction = camera.eye.vector_to(camera.target)
      direction = direction.normalize
      up = camera.up
      fov = camera.fov
      width = model.active_view.vpwidth
      height = model.active_view.vpheight
    
      fovVertical = false
      if(Sketchup.version.to_f >= 15)
        fovVertical = camera.fov_is_height?
      else
        #http://forums.sketchup.com/t/geting-wrong-fov-for-cameras/2197
        fovVertical = camera.aspect_ratio == 0
      end
      
      callCFunction(SUKeyShotExport.method(:add_camera), 'Sketchup Camera', position, direction.to_a, camera.up.to_a, camera.fov, camera.aspect_ratio, width, height, fovVertical)
         
      pages = model.pages
      pages.each do |page| 
        camera = page.camera
        position = camera.eye.to_a
        direction = camera.eye.vector_to(camera.target)
        direction = direction.normalize
        up = camera.up
        fov = camera.fov
        width = model.active_view.vpwidth
        height = model.active_view.vpheight
    
        fovVertical = false
        if(Sketchup.version.to_f >= 15)
          fovVertical = camera.fov_is_height?
        else
          #http://forums.sketchup.com/t/geting-wrong-fov-for-cameras/2197
          fovVertical = camera.aspect_ratio == 0
        end
        callCFunction(SUKeyShotExport.method(:add_camera), page.name, position, direction.to_a, camera.up.to_a, camera.fov, camera.aspect_ratio, width, height, fovVertical)
      end
      
      if cameraOnly
        SUKeyShotExport.disconnect
        model.commit_operation
        return
      end

      idtrans = Geom::Transformation.new

      model.layers.each do |layer|
        if(layer.visible?)
          SUKeyShotExport.begin_group(layer.name)
          model.entities.each do |entity|
            visit(entity, idtrans, nil, layer.name)
          end
          SUKeyShotExport.end_group
        end
      end
    
      if(save)
        SUKeyShotExport.close_file
        result = UI.messagebox "Successfully exported to " + location, MB_OK
      else
        SUKeyShotExport.disconnect
      end
      model.commit_operation
    end # ksexport

    def ks12AnimationExport(location)  
      @error = false    
      begin
        sr = SketchyReplay::SketchyReplay.new()
        sr.start()
      rescue => error
        UI.messagebox(error)
        return
      end
    
      model = Sketchup.active_model
      @texture_set.clear
      @options = Options.new
      callCFunction(SUKeyShotExport.method(:set_options), @options)
      
      # File management
      filename = ""
      unless model.title == ''
        filename += model.title + '.bip'
      else
        filename += 'untitled.bip'
      end
      
      if(location != "")
        save = true
      else
        save = false
      end
      
      callCFunction(SUKeyShotExport.method(:open_file), filename, location, save, false, Sketchup.version)
      if @error
        return
      end
      
      camera = model.active_view.camera

      position = camera.eye.to_a
      direction = camera.eye.vector_to(camera.target)
      direction = direction.normalize
      up = camera.up
      fov = camera.fov
      width = model.active_view.vpwidth
      height = model.active_view.vpheight
    
      fovVertical = false
      if(Sketchup.version.to_f >= 15)
        fovVertical = camera.fov_is_height?
      else
        #http://forums.sketchup.com/t/geting-wrong-fov-for-cameras/2197
        fovVertical = camera.aspect_ratio == 0
      end
      
      callCFunction(SUKeyShotExport.method(:add_camera), 'Sketchup Camera', position, direction.to_a, camera.up.to_a, camera.fov, camera.aspect_ratio, width, height, fovVertical)
      
      sr = SketchyReplay::SketchyReplay.new()
      sr.start()
      frame = 0
      animationMode = 1
      
      callCFunction(SUKeyShotExport.method(:set_duration), sr.lastFrame / 33.0)
      0.upto(sr.lastFrame){
        idtrans = Geom::Transformation.new

        model.layers.each do |layer|
          if(layer.visible?)
            model.entities.each do |entity|
              visitAnimation(entity, idtrans, layer.name)
            end
          end
        end
      
        sr.nextFrame()
        puts frame
        frame += 1
            }
      
      idtrans = Geom::Transformation.new

      model.layers.each do |layer|
        if(layer.visible?)
          SUKeyShotExport.begin_group(layer.name)
          model.entities.each do |entity|
            visit(entity, idtrans, nil, layer.name)
          end
          SUKeyShotExport.end_group
        end
      end
      
      sr.rewind()
      SUKeyShotExport.disconnect
    end  
    
  end #KeyShot12Exporter class

  #
  #   Set up UI hooks
  #

  if not file_loaded?(__FILE__)
    # Create the render and export menu items under Plugins->KeyShot 12
    unless @PLATFORM_IS_OSX and Sketchup.version.to_f < 14
      keyshot12_menu = UI.menu("Plugins").add_submenu("KeyShot 2023")
      ks12_instance = KeyShot12Exporter.new

      if @PLATFORM_IS_OSX
        settings_path = '/Library/Application Support/KeyShot12/KS12Settings.xml'
      else
        if(ENV["KEYSHOT12"])
          settings_path = ENV["KEYSHOT12"] + '/KS12Settings.xml'
        end

        if not File.exists?(settings_path)
          begin
            reg = Win32::Registry::HKEY_CURRENT_USER.open('SOFTWARE\luxion\KeyShot 12', Win32::Registry::KEY_READ | Win32::Registry::KEY_WOW64_64KEY)
            typ, settings_path = reg.read('ResourceFolder')
            settings_path += '/KS12Settings.xml'
          rescue
            begin
              reg = Win32::Registry::HKEY_LOCAL_MACHINE.open('SOFTWARE\luxion\KeyShot 12', Win32::Registry::KEY_READ | Win32::Registry::KEY_WOW64_64KEY)
              typ, settings_path = reg.read('ResourceFolder')
              settings_path += '/KS12Settings.xml'
            rescue
            end
          end
        end
      end

      if File.exists?(settings_path)
        file = File.open(settings_path)
        xml = file.read
        doc = REXML::Document.new(xml)
        s = ""
        doc.elements.each("settings/folders/textures_default") { |element| s = element.text }
        TEXTURE_PATH = s
        
        @@KEYSHOT12_FOUND = true
      end

      if(@@KEYSHOT12_FOUND)
        keyshot12_menu.add_item("Render") do
          ks12_instance.ks12export("", false)
        end
        keyshot12_menu.add_item("Update Camera") do
          ks12_instance.ks12export("", true)
        end
      end

      keyshot12_menu.add_item("Export") do
        filename = ""
        unless Sketchup.active_model.title == ''
          filename += Sketchup.active_model.title + '.bip'
        else
          filename += 'untitled.bip'
        end
        if(ENV["KEYSHOT12"])
          keyshot12path = ENV["KEYSHOT12"] + "\\Scenes\\"
        else
          keyshot12path = ENV["HOME"]
        end
        location = UI.savepanel "Save KeyShot file", keyshot12path, filename
  
        if(location)
          ks12_instance.ks12export(location, false)
        end
      end
      
      keyshot12_menu.add_item("Export SketchyReplay") do
        ks12_instance.ks12AnimationExport("")
      end
      
      keyshot12_menu.add_item("Help") do
        UI.openURL("https://www.keyshot.com/sketchup/")
      end


    # create the toolbar button
    if(@@KEYSHOT12_FOUND)
      renderCmd = UI::Command.new("Render") {
        ks12_instance.ks12export("", false)
      }
      renderCmd.small_icon = "keyshot_icon.png"
      renderCmd.large_icon = "keyshot_icon_large.png"
      renderCmd.tooltip = "Render in KeyShot 2023"
      renderCmd.status_bar_text = "Render in KeyShot 2023"
      renderCmd.menu_text = "Render in KeyShot 2023"
  
      toolbar = UI::Toolbar.new "KeyShot 2023"
      toolbar = toolbar.add_item renderCmd
      toolbar.show
    end
  end
  file_loaded(__FILE__)
end

end #KeyShot12 module
